/* 	SAXFeature.java
 *	Copyright (c) 2006, Brains2B.org
 *	
 *	Created by: Dennis Groenendijk
 *	Created on: 8-jan-2006
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that  * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and 
 * the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products derived from this software 
 * without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.brains2b.saxo;

import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
/** Delegate method for use in SAXReader and SAXWriter
 *  to handle getting and setting features
 *
 * @author <A HREF="MAILTO:dennis@brains2b.nl">dennis@brains2b.nl</A>
 * @version 0.10 [08-01-2006]
 * @since Saxo 0.20
 */
public class SAXFeature {
    
    private String[] m_features;
    private boolean[] m_values;
    
    /**
     * Constructor
     * @param features String[], the URI for the features to add
     * @param values boolean[], the default values for the features
     * @throws ArrayIndexOutOfBoundsException, thrown if the number of features
     * does not match the values
     */
    SAXFeature(String[] features, boolean[] values) {
       if (features.length!=values.length) {
           throw new ArrayIndexOutOfBoundsException("features and default values are not the same length");
       }
       m_features=features;
       m_values=values;
    }
    
    /**
     * get the index of the feature in the feature list
     * @param name String, the fully qualified URI of the feature
     * @return int, the index of the feature or -1 if the feature is
     * not recognized
     */
    private final int feature (String name) {
        for (int i=0;i<m_features.length;i++) {
            if (m_features[i].equals(name)) {
                return i;
            }
        }
        return -1;
    }

    /** 
     * get the value belonging to the index of the feature
     * @param idx, int the index;
     * @return boolean, true if feature set to true, false otherwise
     */
    boolean getValue(int idx) {
        return m_values[idx];
    }

    /** 
     * get the name belonging to the index of the feature
     * @param idx, int the index;
     * @return String, the URI of this feature
     */
    String getName(int idx) {
        return m_features[idx];
    }
 
    /**
     * @see org.xml.sax.XMLReader#getFeature(java.lang.String)
     */
    public boolean getFeature(String name) throws SAXNotRecognizedException,
            SAXNotSupportedException {
        int idx=feature(name);
        if (idx>-1) {
            return m_values[idx];
        } else {
            throw new SAXNotRecognizedException("feature " + name + " is not supported by Saxo");
        }
    }

    /**
     * @see org.xml.sax.XMLReader#setFeature(java.lang.String, boolean)
     */
    public void setFeature(String name, boolean value)
            throws SAXNotRecognizedException, SAXNotSupportedException {
        int idx=feature(name);
        if (idx>-1) {
            m_values[idx]=value;
        } else {
            throw new SAXNotRecognizedException("feature " + name + " is not supported by Saxo");
        }
    }
    
    
}
