/* 	SAXReader.java
 *	Copyright (c) 2002-2006, Brains2B.org
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that  * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and 
 * the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products derived from this software 
 * without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 /* @#History
 * [When]		[Who]					[What]
 * 09-01-2006   dennis@brains2b.nl      Implemented Sax2 specifications
 * TODO Implement LocationImpl instead of position	
 */
package org.brains2b.saxo;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.HashMap;

import org.brains2b.util.XmlHelper;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.DTDHandler;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.AttributesImpl;

/** Partial implementation for SAX 2 Parser
 *  This is a first version of the SAX 2 Parser, 
 *  <B>ContentHandler</B> only.
 *  <p> It is written for to be small, fast and SAX 2 compliant.
 *  <p> Other Parsers will all be stronger in one of these areas, so
 *  if you need only one: small, speed or SAX 2 compliant you 
 *  can probably do better.
 * @author <A HREF="MAILTO:dennis@brains2b.nl">dennis</A>
 * @version 0.20 [09-01-2006]
 */
public class SAXReader implements XMLReader {

	/** an inner class to track the position within the document
	 *   with.
	 * <p>Overwritten toString() to have a readable format for the position
	 */
	protected class Position {
		public int x=0;
		public int y=0;
		
		public String toString() {
			return "(" + x + ", " + y + ")";
		}

	}
    protected final static int SYSTEM=1;
    protected final static int INTERNAL=2;
    
    private final static Attributes EMPTY_ATTS=new AttributesImpl();
    private final static String EMPTY="";
	
    private final SAXFeature m_feature =new SAXFeature(
            new String[] {"http://xml.org/sax/features/namespaces",
            "http://xml.org/sax/features/namespace-prefixes",
            "http://brains2b.org/saxo/features/track-position"},
            new boolean[] {true,true,false}
        );
	private EntityResolver m_resolver;
	private DTDHandler m_dtd;
	private ContentHandler m_content;
	private ErrorHandler m_e_handler;
	private InputSource m_is;
	
	private Position m_pos;
	private BufferedReader m_bfr;
	private HashMap m_namespaces;
    private int m_back= -1;
	
	protected int m_docType;
	protected String m_doc;
		
    public SAXReader() {
        super();
    }
	/**
	 * sets the EntityResolver (this option is not implemented yet)
	 * @see XMLReader#setEntityResolver(EntityResolver)
	 */
	public void setEntityResolver(EntityResolver resolver) {
		m_resolver=resolver;
	}

	/**
	 * sets the DTD Handler (this option is not implemented yet)
	 * @see XMLReader#setDTDHandler(DTDHandler)
	 */
	public void setDTDHandler(DTDHandler dtd) {
		m_dtd=dtd;
	}
	
	/**
	 * set the ContentHandler.
	 * The methods in this interface can be called through this parser
	 * 
	 * @see XMLReader#setContentHandler(ContentHandler)
	 * @since Saxo 0.20
	 */
	public void setContentHandler(ContentHandler handler) {
		m_content=handler;
	}

	/**
	 * set the ErrorHandler (this option is not implemented yet)
	 * @see XMLReader#setErrorHandler(ErrorHandler)
	 */
	public void setErrorHandler(ErrorHandler e_handler) {
		m_e_handler=e_handler;
	}
	
	/* (non-Javadoc)
     * @see org.xml.sax.XMLReader#getErrorHandler()
     */
    public ErrorHandler getErrorHandler() {
        return m_e_handler;
    }

	/**
	 * Main parse function
	 * @see XMLReader#parse(InputSource)
	 */
	public void parse(InputSource is) throws SAXException, IOException {
		if(m_feature.getFeature("http://brains2b.org/saxo/features/track-position")) {
			m_pos=new Position();
		}
		m_is=is;
		m_bfr=getReader();
		readDocument();
		m_bfr.close();
	}

	/** 
	 * Main parse function
	 * @see XMLReader#parse(String)
	 */
	public void parse(String fileName) throws SAXException, IOException {
		if (m_feature.getFeature("http://brains2b.org/saxo/features/track-position")) {
			m_pos=new Position();
		}
        m_namespaces=new HashMap();
		m_is=new InputSource(fileName);
		m_bfr=getReader();
		readDocument();	
		m_bfr.close();	
	}

	/**
	 * top processing level of a document
	 * Will make decision on basic types of input.
	 * @throws SAXException
	 * @throws IOException
	 */
	private void readDocument() throws SAXException, IOException {
		m_namespaces=new HashMap();
        m_content.startDocument();
		
		char c=readChar();
		
		while (c!=(char) -1) {
			if (!XmlHelper.isWhitespace(c)) 
			switch (c) {
				case '<':
					readElement();
					break;
				default:
					backChar(c);
					readContent();
			}
			c=readChar();			
		}
		m_content.endDocument();
		
	}		
	
	/**
	 * read the stream character for character
	 * Is also implemented to return a buffered character if one
	 * is provided.
	 * @return char
	 * @throws IOException
	 */
	private char readChar() throws IOException {
		char c=(char) m_back;
		if (m_back!=-1) {
			m_back=-1;
			return c;
		} else {
			c=(char) m_bfr.read();
			if (m_pos!=null) {
				if (c=='\n') {m_pos.x++; m_pos.y=0;} else {m_pos.y++;} 
			}
		}
		return c;
	}

	/**
	 * backs up one character for next reference to readChar
	 * @see readChar()
	 * @param c
	 */
	private void backChar(char c) {
		m_back=c;
	}

	/**
	 * reads element.
	 * Decides between process instruction, comment and regular element
	 * @throws SAXException
	 * @throws IOException
	 */
	private void readElement() throws SAXException,IOException{
		StringBuffer sb=new StringBuffer();
		Attributes attr=EMPTY_ATTS;
		boolean finished=false;
		boolean start=true;
		char c=readChar();
		switch (c) {
			case '?':
				readProcessInstruction();
				break;
			case '!':
				readComment();
				break;
			case '/':
				finished=true;
				start=false;
				c=readChar();
			default:
				while (c!='/' && c!='>' && !XmlHelper.isWhitespace(c)) {
					sb.append(c);
					c=readChar();
				}
				if (c!='/' && c!='>') {
					attr=readAttributes();
					c=readChar();
				} 
				if (c=='/') {
					readChar(); //Ignore the '>'
					finished=true;
				}
				if (start) {
                    final String name=sb.toString();
					m_content.startElement(nameSpace(name),localName(name),name,attr);	
				}
				if (finished) {
                    final String name=sb.toString();
					m_content.endElement(nameSpace(name),localName(name),name);
				}	
			}
					
				
	}
	
	/**
	 * read everything that isn't markup
	 * @throws SAXException
	 * @throws IOException
	 */
	private void readContent() throws SAXException, IOException {
		StringBuffer sb=new StringBuffer();
		char c=readChar();
		while (c!='<') {
			sb.append(c);
			c=readChar();
		}
		if (c=='<') {
            final String content=sb.toString();
            m_content.characters(content.toCharArray(),0,content.length());
			backChar(c);
		} 
			
	}		
	
	/**
	 * reads attributes from any element, processinstruction and
	 * puts them in an AttributeListImpl.
	 * @see AttributeListImpl
	 * @return AttributeList
	 * @throws IOException
	 */
	private Attributes readAttributes() throws IOException,SAXException {
		AttributesImpl attr=new AttributesImpl();
		StringBuffer buf=new StringBuffer();
		String name=null;
		boolean inLiteral=false;
		char c=readChar() ;
		while ((c!='/' && c!='>') || inLiteral) {
			if (XmlHelper.isWhitespace(c) && !inLiteral && name!=null) {
				if (isNamespace(name)) {
                    addNamespace(name,buf.toString());
                } else {
                    attr.addAttribute(nameSpace(name),localName(name),name,"CDATA",buf.toString());
                }
				buf.delete(0,buf.length());
				name=null;
			} else if (c=='='&& !inLiteral) {
				name=buf.toString().trim();
                buf.delete(0,buf.length());
			} else if (c=='\"' || c=='\'') {
				inLiteral=!inLiteral;
			} else {
				buf.append(c);
			}
			c=readChar();
		}
		if (name!=null) {
            if (isNamespace(name)) {
                addNamespace(name,buf.toString());
            } else {
                attr.addAttribute(nameSpace(name),localName(name),name,"CDATA",buf.toString());
            }
		}
		backChar(c);
		return attr;
	}
	
	/**
	 * reads ProcessInstruction
	 * This function will ignore processing instructions for the moment.
	 * TODO keep track of attributes given through the processing instructions
	 * @throws IOException
	 */
	private void readProcessInstruction() throws IOException {
		char c=readChar();
		while (c!='?') {
			c=readChar();
		}
		c=readChar(); //ignore >
	}

	/**
	 * read comments and !DOCTYPE tags. 
	 * Also add ![CDATA which will change in next versions
	 * TODO implement processing for &lt![CDATA[&gt
	 * @throws IOException
	 * @throws SAXException
	 */
	private void readComment() throws IOException, SAXException {
		char c=readChar();
		switch (c) {
			case '-':
				if (readChar()=='-') {
					c=readChar();
					while (true) {
						if (c=='-' && readChar()=='-' && readChar()=='>') {
							break;
						}
						c=readChar();
					}
				}
				break;
			case 'D':
				if (readChar()=='O' && readChar()=='C' && readChar()=='T' && readChar()=='Y' && readChar()=='P' && readChar()=='E') {
					c=readChar();
					StringBuffer sb=new StringBuffer();
					boolean open=false;
					while (true) {
						if (c=='S') {
							if (readChar()=='Y' && readChar()=='S' && readChar()=='T' && readChar()=='E' && readChar()=='M') {
								m_docType=SYSTEM;
								open=true;
							}
						} else if (c=='[') {
							m_docType=INTERNAL;
							open=true;
						} else if (c=='\'' || c=='\"') {
							//ignore quotes
						} else if (open && c!=']' && c!='>') {
							sb.append(c);
						} else if (c=='>') {
							m_doc=sb.toString().trim();
							break;
						}
						c=readChar();
					}
					
				}
				break;
			default:
				while (c!='>') {
						c=readChar();
					}			
		}
	}
	
	/**
	 * gets a BufferedReader from the arguments given through
	 * the parser function.
	 * TODO needs to check for URL and File
	 * @return BufferedReader
	 * @throws IOException
	 */
	private BufferedReader getReader() throws IOException{
		if (m_is.getByteStream()!=null) {
			return new BufferedReader(new InputStreamReader(m_is.getByteStream()));
		} else if (m_is.getCharacterStream()!=null) {
			return new BufferedReader(m_is.getCharacterStream());
		} else if (m_is.getSystemId()!=null) {
			try {
				File fl=new File(m_is.getSystemId());
				return new BufferedReader(new FileReader(fl));
			} catch (FileNotFoundException ex) {
				return new BufferedReader(new InputStreamReader(new URL(m_is.getSystemId()).openStream()));
			}
		}
		throw new IOException("File could not be read");
	}		
	
	/**
	 * return the current position of the reader within the document
	 * <p>Easiest way to use is just call toString() on position to have
	 * a printable format for the position
	 * @return SAXParser.Position
	 * @see SAXReader.Position
	 */
	protected Position getPosition() {
		return m_pos;
	}
	/**
	 * get the content handler for this parser
	 * @since Saxo 0.20
	 */
    public ContentHandler getContentHandler() {
        return m_content;
    }
    
    /**
	 * get the dtd for this parser
	 * @since Saxo 0.20
	 */
    public DTDHandler getDTDHandler() {
        return m_dtd;
    }
    
    /**
	 * get the entity resolver for this parser
	 * @since Saxo 0.20
	 */
    public EntityResolver getEntityResolver() {
        return m_resolver;
    }
    
    public boolean getFeature(String name) throws SAXNotRecognizedException,
            SAXNotSupportedException {
        return m_feature.getFeature(name);
    }
    
    public Object getProperty(String name) throws SAXNotRecognizedException,
            SAXNotSupportedException {
       throw new SAXNotRecognizedException("property " + name + " is not supported by Saxo");
    }
    
    public void setFeature(String name, boolean value)
            throws SAXNotRecognizedException, SAXNotSupportedException {
       m_feature.setFeature(name,value);
    }
    
    public void setProperty(String name, Object value)
            throws SAXNotRecognizedException, SAXNotSupportedException {
        throw new SAXNotRecognizedException("property " + name + " is not supported by Saxo");
    }
    
    /**
     * return the localName from a qualified name, or the part after ':'
     * @param s String, the qualified name, pre:localName
     * @return String, the local name, localName
     */
    private final String localName(String s) {
        final int idx=s.indexOf(':');
        if (idx>-1) {
            return s.substring(idx+1);
        } else {
            return s;
        }
    }
    
    /**
     * check if the qualified name is the definition of a namespace
     * @param s String, the qualified name to check
     * @return boolean, true if name starts with xmlns, false otherwise
     */
    private final boolean isNamespace(String name) {
        if (name!=null && name.startsWith("xmlns")) {
            return true;
        }
        return false;
    }

    /**
     * return a namespace URI from a qualified name, by looking
     * it up in the prefix-resolver
     * @param s String, the qualified name, 'pre:localName'
     * @return String, the namespace URI, 'http://brains2b.org/saxo'
     */
    private final String nameSpace(String name) {
       final int idx=name.indexOf(':');
       if (idx>-1) {
           return (String) m_namespaces.get(name.substring(0,idx));
       } else {
           return EMPTY;
       }
    }
    
    /** add a namespace to the prefix-resolver
     * @param name String, the qualified namespace, 'xmlns:pre'
     * @param value String, the namespace URI, 'http://brains2b.org/saxo'
     * @throws SAXException, thrown if ContentHandler.startPrefixMapping fails
     * @see ContentHandler#startPrefixMapping(java.lang.String, java.lang.String)
     */
    private final void addNamespace(String name, String value) throws SAXException {
        final String prefix=localName(name); 
        m_namespaces.put(prefix,value);
        m_content.startPrefixMapping(prefix,value);
    }
}
