/* 	SAXWriter.java
 *	Copyright (c) 2002-2006, Brains2B.org
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that  * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and 
 * the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products derived from this software 
 * without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 /* @#History
 * [When]		[Who]				[What]
 * 08/26/2002	dennis@brains2b.nl  Write terminator on same line if content only,
 * 									corrected indent for elements without content
 * 11/17/2002		"		"		Added possibility to add Comment
 * 09-01-2006   dennis@brains2b.nl  Implemented SAX2 interface methods
 */
package org.brains2b.saxo;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.HashMap;
import java.util.Iterator;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.helpers.AttributesImpl;

/** A SAX method like XML writer
 *  To be sure there is <b>no</b> such thing as a SAXWriter for SAX 2, 
 *  in as much as it isn't defined in the API. 
 *  
 * <p>This class is provided because I needed it and because I think
 *  it would be handy to have a Writer that takes at least part
 *  of the DocumentHandler interface not to read but to write a document
 *  preventing me to have to spell out all the tags every time I need to write XML data to file.
 * 
 * <p> Public Members throw SAXException, instead of IOException to keep
 * in line with the SAX specification. Possible next versions will throw
 * subtypes of SAXException, containing writing information and position. 
 * 
 * <p><i>Note:</i> The Writer writes to files only if you supply it by String
 * no URL connections will be openend. If you want to write to another target use
 * the SAXWriter(Writer) constructor instead.
 * 
 * @author <A HREF="MAILTO:dennis@brains2b.nl">dennis</A>
 * @version 0.20 [09-01-2006]
 */
public class SAXWriter {
	
	private final static int START=0;
	private final static int END=1;
	private final static int BOTH=2;
	private int m_fmtPos;
    private final SAXFeature m_feature=new SAXFeature(
               new String[] {"http://brains2b.org/saxo/features/format-write"},
               new boolean[] {true}
               );
	private Writer m_writer;
	private String m_location;
	private boolean m_contentOnly;
    
    private HashMap m_prefix;
    private AttributesImpl m_namespaces;
	
	/** 
	 * Default empty constructor. 
	 * Don't forget to set Location or Writer through
	 * the setters.
	 * 
	 */
	public SAXWriter() {
	   super();
    }
	
	/**
	 * Constructor
	 * Sets the Writer where the output will go.
	 * @param writer java.io.Writer
	 */
	public SAXWriter(Writer writer) {
		m_writer=writer;
	}
	
	/**
	 * Constructor
	 * Sets the file to open a StreamWriter on.
	 * Opening of the stream will not happen until startDocument()
	 * is called.
	 * @param location String
	 */
	public SAXWriter(String location) {
		m_location=location;	
	}
	
	/**
	 * start Document.
	 * starts the document and will open de StreamWriter if it
	 * is not alreay open.
	 * <p> Does not insert the default processing Instruction for
	 * XML. You have to call insertDescriptor() yourself.
	 * @see #insertDescriptor()
	 * @throws SAXException
	 */
	public void startDocument() throws SAXException {
		try {
		m_prefix=new HashMap();
        if (m_writer==null) {
			File fl=new File(m_location);
			m_writer=new FileWriter(fl);
		} 
		} catch (Exception ex) {
			throw new SAXException(ex);
		}
		
	}
	
	/**
	 * ends the Document.
	 * This will end the document and flushes and closes the Stream
	 * to the document. This method should be called once and precisely
	 * once after you called startDocument to free up resources.
	 * @throws SAXException
	 */
	public void endDocument() throws SAXException {
		try {
			m_writer.flush();
			m_writer.close();
		} catch (Exception ex) {
			throw new SAXException(ex);
		}
	}
	
	/**
	 * insert Descriptor.
	 * <p>Inserts a descriptor of the type:
	 * &ltxml version=&quot1.0&quot encoding=&quotUTF-8&quot &gt
	 * into the document. If you need it. If you need it call it 
	 * after startDocument to end up with a valid xml document.</p>
	 * 
	 * @throws SAXException
	 */
	public void insertDescriptor() throws SAXException {
		AttributesImpl attr=new AttributesImpl();
		//should be configurable, but for the moment it will do
		attr.addAttribute(null,null,"version","CDATA","1.0");
		attr.addAttribute(null,null,"encoding","CDATA","UTF-8"); 
		try {
			m_writer.write("<?xml");
			m_writer.write(' ');
			writeAttributes(attr);
			m_writer.write("?>");
		} catch (Exception ex) {
			throw new SAXException(ex);
		}
		
	}
	
    /**
     * @see ContentHandler#startPrefixMapping(String, String)
     */
    public void startPrefixMapping(String prefix, String uri) throws SAXException {
        addNamespace(prefix,uri);
    }
    
    /**
     * @see ContentHandler#endPrefixMapping(String)
     */
    public void endPrefixMapping(String prefix) throws SAXException {
        removeNamespace(prefix);
    }

    /**
     * start a comment in an XML document
     * <p>any content or element written after calling
     * start en before calling end will be written inside
     * the comment
     * @throws SAXException if start cannot be written
     */
	public void startComment() throws SAXException {
		try {
			writeFormat(START);
			m_writer.write("<!--");
		} catch (IOException ex) {
			throw new SAXException(ex);
		}
	}
	
    /**
     * end a comment in an XML document
     * <p>any content or element written after calling
     * start en before calling end will be written inside
     * the comment
     * @throws SAXException if end cannot be written
     */
    public void endComment() throws SAXException {
        try {
            writeFormat(END);
            m_writer.write("-->");
        } catch (IOException ex) {
            throw new SAXException(ex);
        }
    }
    
	/**
	 * start Element.
	 * This is the start of an element that has Content.
	 * <p>For EMPTY elements use insertElement.
	 * <UL>Correct sequence for non-EMPTY elements is
	 * <LI>startElement(String, AttributeList)
	 * <LI>setContent(String) <b>or</b> insert other Elements
	 * <LI>endElement(String)
	 * </UL>
	 * <p>You can insert <i>null</i> for the AttributeList if you don't need
	 * it. Parsers however will always return an AttributeList Object
	 * even it contains no attributes.
	 * @see #insertElement(String, Attributes)
	 * @param name
	 * @param attr 
	 * @throws IOException
	 */
	public void startElement(String name, Attributes attr) throws SAXException {
		startElement(null,null,name,attr);
	}

    /**
     * start Element.
     * This is the start of an element that has Content.
     * <p>For EMPTY elements use insertElement.
     * <UL>Correct sequence for non-EMPTY elements is
     * <LI>startElement(String, AttributeList)
     * <LI>setContent(String) <b>or</b> insert other Elements
     * <LI>endElement(String)
     * </UL>
     * <p>You can insert <i>null</i> for the AttributeList if you don't need
     * it. Parsers however will always return an AttributeList Object
     * even it contains no attributes.
     * @see #insertElement(String, Attributes)
     * @param namespaceURI String, the namespace to declare
     * @param localname String
     * @param qname String
     * @param attr Attributes
     * @throws SAXException
     */
    public void startElement(String namespaceURI, String localname, String qname, Attributes attr) throws SAXException {
       if (qname==null) {
            qname=resolveNamespace(namespaceURI,localname);
        }
        try {
            writeFormat(START);
            m_writer.write("<");
            m_writer.write(qname);
            if (m_namespaces!=null) {
                writeAttributes(m_namespaces);
                m_namespaces=null;
            }
            writeAttributes(attr);
            m_writer.write(">");
        } catch (IOException ex) {
            throw new SAXException(ex);
        }
    }
    
	/**
	 * end Element.
	 * Ends a non-EMPTY element
	 * <p>For EMPTY elements use insertElement.
	 * <p>Do not call this after insertElement.
	 * @see #insertElement(String, Attributes)
	 * @param name String, the local or qname of the element
	 * @throws SAXException
	 */
	public void endElement(String name) throws SAXException {
        endElement(null,null,name);
    }
    /**
     * end Element.
     * Ends a non-EMPTY element
     * <p>For EMPTY elements use insertElement.
     * <p>Do not call this after insertElement.
     * @see #insertElement(String, Attributes)
     * @param namespaceURI String
     * @param localname String, the localname the element
     * @param qname String, the qname of the element
     * @throws SAXException
     */
    public void endElement(String namespaceURI, String localname, String qname) throws SAXException {
		if (qname==null) {
            qname=resolveNamespace(namespaceURI,localname);
        }
        try {
			writeFormat(END);
			m_writer.write("</");
			m_writer.write(qname);
			m_writer.write(">");
		} catch (IOException ex) {
			throw new SAXException(ex);
		}
	}

	/**
	 * inserts an EMPTY Element
	 * <p>Although I would not see a purpose for this method if you do, you 
	 * can leave the AttributeList empty.
	 * For non-EMPTY Elements use startElement.</p>
	 * @see #startElement(String, Attributes)
	 * @param name String, can be either a localname or qname
	 * @param attr Attributes
	 * @throws SAXException
	 */
	public void insertElement(String name, Attributes attr) throws SAXException {
		insertElement(null,null,name,attr);
	}
    
    /**
     * inserts an EMPTY Element
     * <p>For non-EMPTY Elements use startElement.</p>
     * @see #startElement(String, Attributes)
     * @param namespaceURI String, the namespace to declare
     * @param localname String, can be either a localname or qname
     * @param qname String, null if setting the default namespace
     * @param attr Attributes, null if no attributes need to be defined
     * @throws SAXException
     */
    public void insertElement(String namespaceURI,String localname,String qname,Attributes attr) throws SAXException {
        if (qname==null) {
            qname=resolveNamespace(namespaceURI,localname);
        }
        try {
            writeFormat(BOTH);
            m_writer.write("<");
            m_writer.write(qname);
            if (m_namespaces!=null) {
                writeAttributes(m_namespaces);
                m_namespaces=null;
            }
            writeAttributes(attr);
            m_writer.write("/>");
        } catch (IOException ex) {
            throw new SAXException(ex);
        }
    }
    
    
	/**
	 * sets the Character Content of an Element. 
	 * There is no protection against writing outside 
	 * Element tags yet, so you have to build that protection
	 * into the code that uses this.
	 * @param content
	 * @throws SAXException
	 */
	public void setContent(String content) throws SAXException{
		try {
			m_writer.write(content);
			m_contentOnly=true;
		} catch (IOException ex) {
			throw new SAXException(ex);
		}
	}

    public boolean getFeature(String name) throws SAXNotRecognizedException,
            SAXNotSupportedException {
        return m_feature.getFeature(name);
    }

    public void setFeature(String name, boolean value)
            throws SAXNotRecognizedException, SAXNotSupportedException {
        m_feature.setFeature(name,value);  
    }

    
	/**
	 * gets the Writer used for output
	 * @return Writer
	 */
	public Writer getWriter() {
		return m_writer;
	}

	/**
	 * sets the Writer used for output
	 * @param writer
	 */
	public void setWriter(Writer writer) {
		m_writer = writer;
	}

    private void addNamespace(String prefix, String namespace) {
        String pf="xmlns";
        if (prefix!=null) {
            pf="xmlns:"+prefix;
        }
        if (m_namespaces==null) m_namespaces= new AttributesImpl();
        m_namespaces.addAttribute(null,prefix,pf,"CDATA",namespace);
        m_prefix.put(namespace,prefix);
    }
    
    private void removeNamespace(String prefix) {
        if (prefix==null || !m_prefix.containsValue(prefix)) return;
        for (Iterator it=m_prefix.keySet().iterator();it.hasNext();) {
            String key = (String) it.next();
            if (prefix.equals(m_prefix.get(key))) {
                m_prefix.remove(key);
                break;
            }
        }
    }
    
    private String resolveNamespace(String namespaceURI,String localname) {
        String prefix= (String) m_prefix.get(namespaceURI);
        if (prefix!=null) {
            return prefix +":" + localname;
        } else {
            return localname;
        }
    }
    
    /**
     * Calculates the number of tabs given by the depth of the
     * element
     * @return String
     */
    private String getTabs() {
        StringBuffer sb=new StringBuffer();
        for (int i=0;i<m_fmtPos;i++) {
            sb.append('\t');
        }
        return sb.toString();
    }
    
	/**
	 * writes Attributes into the element tags.
	 * Only supports CDATA at the moment.
	 * @param attr
	 * @throws IOException
	 */
	private void writeAttributes(Attributes attr) throws IOException {
		if (attr==null) return;
		for (int i=0; i<attr.getLength();i++) {
            String qName=null;
            if (attr.getQName(i)==null && attr.getURI(i)!=null) {
                qName=(String) m_prefix.get(attr.getURI(i));
                qName+=":"+attr.getLocalName(i);
            } else {
                qName=attr.getQName(i);
            }
			m_writer.write(' ');
			m_writer.write(qName);
			m_writer.write("=");
			if(attr.getType(i)=="CDATA") { 
				m_writer.write("\"");
				m_writer.write(attr.getValue(i));
				m_writer.write("\"");
			} else {
				m_writer.write(attr.getValue(i));
			}
			
		}
	}
	
	/**
	 * writes the Formatting 
	 * Formatting is done by tabs (configuration will follow later). 
	 * Indenting is done after each startElement and starts with a
	 * carriage return. The field m_fmtPos keeps track of the dept of the element
	 * @param pos The position of the Element written after the formatting. Can be: START,END or BOTH
	 * @throws IOException
	 */
	private void writeFormat(int pos) throws IOException {
		switch (pos) {
			case START:
				m_writer.write("\n");
				m_writer.write(getTabs());
				m_fmtPos++;
				break;
			case END:
				m_fmtPos--;
				if (!m_contentOnly) {
					m_writer.write("\n");
					m_writer.write(getTabs());
				}
				m_contentOnly=false;
				break;
			case BOTH:
				m_writer.write("\n");
				m_writer.write(getTabs());
			default:
		}
	}
	    
}