/* 	SAXWriterTest.java
 *	Copyright (c) 2006, Brains2B.org
 *	
 *	Created by: Dennis Groenendijk
 *	Created on: 9-jan-2006
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that  * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and 
 * the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products derived from this software 
 * without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package test.org.brains2b.saxo;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;

import junit.framework.TestCase;

import org.brains2b.saxo.SAXReader;
import org.brains2b.saxo.SAXWriter;
import org.brains2b.util.XmlHelper;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;

/** JUnit test for org.brains2b.saxo.SAXWriter
 *
 * @author <A HREF="MAILTO:dennis@brains2b.nl">dennis@brains2b.nl</A>
 * @version 0.10 [09-01-2006]
 */
public class SAXWriterTest extends TestCase {
    private class CloneXMLHandler implements ContentHandler {
        private SAXWriter m_writer;
        public CloneXMLHandler(SAXWriter writer) {
            super();
            m_writer=writer;
        }
        
        public void characters(char[] ch, int start, int length) throws SAXException {
            m_writer.setContent(new String(ch,start,length)); 
        }

        public void endDocument() throws SAXException {
            m_writer.endDocument();
        }

        public void endElement(String namespaceURI, String localName, String qName) throws SAXException {
            m_writer.endElement(qName);
        }
        

        public void startDocument() throws SAXException {
            m_writer.startDocument();
            m_writer.insertDescriptor();
            
        }

        public void startElement(String namespaceURI, String localName, String qName, Attributes atts) throws SAXException {
            m_writer.startElement(namespaceURI,localName,qName,atts);
        }

        public void startPrefixMapping(String prefix, String uri) throws SAXException {
           m_writer.startPrefixMapping(prefix,uri);
            
        }
        
        public void endPrefixMapping(String prefix) throws SAXException {
            m_writer.endPrefixMapping(prefix);
        }

        //Not implemented for this test
        public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException {
        }

        public void processingInstruction(String target, String data) throws SAXException {
        }

        public void setDocumentLocator(Locator locator) {
        }

        public void skippedEntity(String name) throws SAXException {
        }
        
    }
    
    public static void main(String[] args) {
        junit.swingui.TestRunner.run(SAXWriterTest.class);
    }
    
    public SAXWriterTest(String name) {
        super(name);
    }
    
    public void testSetFeatures() {
        SAXWriter sax=new SAXWriter();
        final String feature="http://brains2b.org/saxo/features/format-write"; 
        try {
            assertTrue(sax.getFeature(feature)); //the default
            sax.setFeature(feature,false);
            assertTrue(!sax.getFeature(feature)); 
        } catch (SAXException saex) {
            assertTrue(saex.getMessage(),false);
        }
    }
    
    public void testFeatureException() {
        SAXWriter sax=new SAXWriter();
        try {
            assertTrue(!sax.getFeature("http://brains2b.org/saxo/this-is-not-a-feature"));
        } catch (SAXNotRecognizedException snrex) {
            assertTrue(true);
        } catch (SAXException saex) {
            assertTrue(saex.getMessage(),false);
        }
    }
    
    public void testCloneCopy() {
        
        try {
            File temp=File.createTempFile("junit",".xml");    
            //temp.deleteOnExit();
            File in =new File("test/result/writetest.xml");
            FileWriter fw=new FileWriter(temp);
            SAXWriter sax=new SAXWriter(fw);
            SAXReader reader=new SAXReader();
            
            reader.setContentHandler(new CloneXMLHandler(sax));
            reader.parse(new InputSource(new FileReader(in)));
            
            compareFiles(in,temp);
            
        } catch (IOException iex) {
            assertTrue(iex.getMessage(),false);
        } catch (SAXException saex) {
            assertTrue(saex.getMessage(),false);
        }
    }
    
    public void testPrefix() {
        StringWriter sw=new StringWriter();
        SAXWriter sax=new SAXWriter(sw);
        try {
            sax.startDocument();
            sax.startPrefixMapping("x","urn:my-uri");
            sax.startElement("urn:my-uri","element",null,null);
            sax.insertElement("urn:my-uri","sub-element",null,null);
            sax.endElement("x:element");
            sax.endDocument();
            assertEquals("\n<x:element xmlns:x=\"urn:my-uri\">\n\t<x:sub-element/>\n</x:element>",sw.toString());
            
        } catch (SAXException saex) {
            assertTrue(saex.getMessage(),false);
        }
        
    }
    
    public void testDefaultNamespace() {
        StringWriter sw=new StringWriter();
        SAXWriter sax=new SAXWriter(sw);
        try {
            sax.startDocument();
            sax.startPrefixMapping(null,"urn:my-uri");
            sax.startElement("urn:my-uri","element",null,null);
            sax.endElement("urn:my-uri","element",null);
            sax.endDocument();
            assertEquals("\n<element xmlns=\"urn:my-uri\">\n</element>",sw.toString());
        } catch (SAXException saex) {
            assertTrue(saex.getMessage(),false);
        }
        
    }
    
    private void compareFiles(File in, File temp) throws IOException {
        FileReader rin=new FileReader(in);
        FileReader rtemp=new FileReader(temp);
        char[] bufin=new char[1024];
        char[] buftemp=new char[1024];
        int len=0;
        int total=0;
        while ((len=rin.read(bufin))!=-1) {
            rtemp.read(buftemp);
            for (int i=0;i<len;i++) {
                if (!XmlHelper.isWhitespace(bufin[i])) 
                assertEquals("character at :" + (total+i) + "differs",bufin[i],buftemp[i]);
            }
            total+=len;
        }
    }
}
